from __future__ import annotations

from copy import copy
from inspect import isclass
from inspect import isfunction
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

from slightly_better_types.function import Function


class Handlers:

    def __init__(self, _class):
        self._class = _class
        self.methods = {
            func_name: Function(
                getattr(self._class, func_name)
            ) for func_name in dir(self._class)
            if isfunction(getattr(self._class, func_name))
        }
        self.filters = []
        self.visibility_filters = []

    @staticmethod
    def new(class_or_class_instance: Any):
        if not isclass(class_or_class_instance):
            _class = type(class_or_class_instance)
        else:
            _class = class_or_class_instance
        return Handlers(_class)

    def accepts(self, *args, **kwargs) -> Handlers:
        func: Callable[[Function], bool] = lambda method: method.accepts(*args, **kwargs)
        return self.filter(func)

    def all(self) -> Dict[str, Function]:
        all_methods = {}
        for func_name, func in self.methods.items():
            if not self._filter_allows(method=func):
                continue

            all_methods[func_name] = func
        return all_methods

    def filter(self, _filter: Callable[..., bool]) -> Handlers:
        clone = copy(self)
        clone.filters = self.filters[:]
        clone.filters.append(_filter)
        return clone

    def first(self) -> Optional[Function]:
        return next(iter(self.all().values()), None)

    def public(self) -> Handlers:
        clone = copy(self)
        clone.visibility_filters = self.visibility_filters[:]
        clone.visibility_filters = Function.PUBLIC
        return clone

    def protected(self) -> Handlers:
        clone = copy(self)
        clone.visibility_filters = self.visibility_filters[:]
        clone.visibility_filters = Function.PROTECTED
        return clone

    def private(self) -> Handlers:
        clone = copy(self)
        clone.visibility_filters = self.visibility_filters[:]
        clone.visibility_filters = Function.PRIVATE
        return clone

    def _filter_allows(self, method: Function) -> bool:
        if self.visibility_filters and method.visibility() not in self.visibility_filters:
            return False

        for _filter in self.filters:
            if _filter(method) is False:
                return False

        return True



from inspect import Parameter as InspectParameter
from inspect import signature
from typing import Any
from typing import get_type_hints
from typing import List

from slightly_better_types.parameter import Parameter
from slightly_better_types.utils import inspect_type


class Function:
    PUBLIC = 'public'
    PROTECTED = 'protected'
    PRIVATE = 'private'

    def __init__(self, _function):
        self._function = _function

        self.parameters = {
            InspectParameter.KEYWORD_ONLY: [],
            InspectParameter.VAR_KEYWORD: [],
            InspectParameter.VAR_POSITIONAL: [],
            InspectParameter.POSITIONAL_ONLY: [],
            InspectParameter.POSITIONAL_OR_KEYWORD: []
        }

        # signature returns whatever is in the function's __annotation__ attribute
        # which could get distorted if `__future__.annotations` is used anywhere
        self._signature = signature(self._function)

        type_hints = get_type_hints(self._function)
        parameter_list = list(self._signature.parameters.values())
        self.kinds = []

        sb_parameters = [
            Parameter.new(
                parameter=parameter,
                type_hint=type_hints.get(parameter.name)
            ) for parameter in parameter_list
        ]

        for sb_parameter in sb_parameters:
            self.parameters[sb_parameter.kind].append(sb_parameter)
            self.kinds.append(sb_parameter.kind)

        self.input_count = len(self.parameters.values())

    def accepts(self, *args: Any, **kwargs) -> bool:
        try:
            param_names = [param.name for param in self.parameters[InspectParameter.POSITIONAL_OR_KEYWORD]]
            temp_args = args
            if 'self' in param_names or 'cls' in param_names:
                temp_args = [None, *args]
            self._signature.bind(*temp_args, **kwargs)
        except TypeError:
            return False

        positionally_accepting_params: List[Parameter] = [
            param for param in [
                *self.parameters[InspectParameter.POSITIONAL_ONLY],
                *self.parameters[InspectParameter.POSITIONAL_OR_KEYWORD],
                *self.parameters[InspectParameter.VAR_POSITIONAL],
            ] if param.name not in ['self', 'cls']
        ]

        for index, arg in enumerate(args):
            try:
                current_param = positionally_accepting_params[index]
                if not current_param.accepts(arg):
                    return False
            except IndexError:
                if (
                        positionally_accepting_params[-1].kind == InspectParameter.VAR_POSITIONAL
                        and positionally_accepting_params[-1].accepts(arg)
                ):
                    continue
                else:
                    return False

        kwarg_accepting_params = {}
        for param in [
            *self.parameters[InspectParameter.POSITIONAL_OR_KEYWORD],
            *self.parameters[InspectParameter.KEYWORD_ONLY],
            *self.parameters[InspectParameter.VAR_KEYWORD],
        ]:
            kwarg_accepting_params[param.name] = param

        for key, value in kwargs.items():
            if key not in kwarg_accepting_params:
                if (
                        self.parameters[InspectParameter.VAR_KEYWORD]
                        and not self.parameters[InspectParameter.VAR_KEYWORD][0].accepts(value)
                ):
                    return False
            elif not kwarg_accepting_params[key].accepts(value):
                return False
        return True

    def get_name(self) -> str:
        return self._function.__name__

    def visibility(self) -> str:
        name = self.get_name()
        if name.startswith('__'):
            return self.PRIVATE
        elif name.startswith('_'):
            return self.PROTECTED
        else:
            return self.PUBLIC

    def is_static(self):
        return inspect_type(self._function) == 'staticmethod'

    def is_abstract(self):
        return hasattr(self._function, '__isabstractmethod__')

    def is_public(self):
        return not self.get_name().startswith('_')

    def is_protected(self):
        return self.get_name().startswith('_') and not self.get_name().startswith('__')

    def is_private(self):
        return self.get_name().startswith('__')



import unittest
from typing import Dict

from slightly_better_types.handlers import Handlers


class Bar:

    def accepts_string(self, a: str):
        pass

    def accepts_string_too(self, b: str):
        pass

    def accepts_int(self, c: int):
        pass

    def _protected(self, a: Dict):
        pass

    def __private(self, a: Dict):
        pass

    @staticmethod
    def static(a: float):
        pass


class TestHandlers(unittest.TestCase):

    def test_find(self):
        handlers = Handlers(Bar)

        assert ['accepts_string', 'accepts_string_too'] == list(handlers.accepts('string').all().keys())
        assert ['accepts_string_too'] == list(handlers.accepts(b='string').all().keys())
        print(list(handlers.accepts(1).all().keys()))
        assert ['accepts_int', 'static'] == list(handlers.accepts(1).all().keys())

        class TestClass:
            pass

        assert [] == list(handlers.accepts(TestClass()).all().keys())

    def test_first(self):
        handlers = Handlers(Bar)
        assert 'accepts_string' == handlers.accepts('string').first().get_name()
        assert 'accepts_string_too' == handlers.accepts(b='string').first().get_name()

        class TestClass:
            pass

        assert None is handlers.accepts(TestClass()).first()

    def test_visibility(self):
        assert None is Handlers.new(Bar).public().accepts([]).first()
        assert None is Handlers.new(Bar).protected().accepts([]).first()
        assert Handlers.new(Bar).private().accepts({}).first() is not None
        assert Handlers.new(Bar).public().protected().accepts({}).first() is not None

    def test_all(self):
        assert 1 == len(Handlers.new(Bar).private().all())
        assert 4 == len(Handlers.new(Bar).public().all())
        assert 6 == len(Handlers.new(Bar).all())
        assert 1 == len(Handlers.new(Bar).filter(lambda x: x.is_static()).all())

    def test_filter(self):
        assert Handlers.new(Bar).filter(lambda x: not x.is_static()).accepts(2.1).first() is None
        assert Handlers.new(Bar).filter(lambda x: x.is_static()).accepts(1.2).first() is not None


if __name__ == '__main__':
    unittest.main()



import sys


def inspect_type(func):
    if '.' not in func.__qualname__:
        ftype = 'function'
    else:
        # __qualname__: 'className.functioNname'
        cls_name = func.__qualname__.rsplit('.', 1)[0]
        # Get the class by name
        cls = getattr(sys.modules[func.__module__], cls_name)
        # cls.__dict__[func.__name__] should return like <class 'staticmethod'>
        func_name = func.__name__
        if func.__name__.startswith('__'):
            func_name = f"_{cls_name}{func.__name__}"

        ftype = cls.__dict__[func_name].__class__.__name__
        if ftype not in ['staticmethod', 'classmethod', 'function']:
            raise TypeError('Unknown Type %s, Please check input is method or function' % func)
    return ftype



import inspect
import unittest
from abc import ABC
from types import FunctionType
from typing import Any  # noqa
from typing import Dict  # noqa
from typing import get_type_hints
from typing import List  # noqa
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union  # noqa

from slightly_better_types.parameter import Parameter

if TYPE_CHECKING:
    from typing import Generator


class AbstractFoo(ABC):
    pass


class Foo(AbstractFoo):
    def __init__(self):
        pass


class TestSlightlyBetterParameter(unittest.TestCase):
    boolean: bool = True
    integer: int = 1
    _float: float = 1.2
    string: str = 'string'
    _none = None
    foo: Foo

    def true_values(self) -> "Generator":

        self.foo = Foo()

        yield ['a: dict', {self.string: self.integer}]

        yield ['a: Any', self.boolean]
        yield ['a: bool', self.boolean]
        yield ['a: Union[int, bool]', self.boolean]

        yield ['a: Any', self._float]
        yield ['a: float', self._float]
        yield ['a: Union[int, float]', self._float]

        yield ['a: Any', self.integer]
        yield ['a: int', self.integer]
        yield ['a: Union[int, str]', self.integer]

        yield ['a: Any', self.string]
        yield ['a: str', self.string]
        yield ['a: Union[int, str]', self.string]

        yield ['a: Any', [self.string]]
        yield ['a: List[str]', [self.string]]
        yield ['a: Union[int, List[str]]', [self.string]]

        yield ['a: Any', self._none]
        yield ['a: Optional[str]', self._none]
        yield ['a: Union[int, Optional[str]]', self._none]

        yield ['a: Any', self.foo]
        yield ['a: Optional[Foo]', self.foo]
        yield ['a: AbstractFoo', self.foo]

    def false_values(self):
        yield ['a: Union[str, int]', []]
        yield ['a: Union[str, int]', None]
        yield ['a: int', 'invalid']
        yield ['a: int', 2.1]

    def test_true(self):
        for definition, arg in self.true_values():
            assert self._make_type(definition).accepts(arg)

    def test_false(self):
        for definition, arg in self.false_values():
            assert not self._make_type(definition).accepts(arg)

    def _make_type(self, definition: str) -> Optional[Parameter]:
        exec(f"class TestClass:\n\tdef test(self,{definition}):\n\t\tpass")
        func: FunctionType = (eval('TestClass.test'))
        param = list(inspect.signature(func).parameters.values())[1]
        if param:
            type_hints = get_type_hints(func)
            return Parameter(
                name=param.name,
                kind=param.kind,
                default=param.default,
                annotation=param.annotation,
                type_hint=type_hints.get(param.name)
            )
        return None



import unittest
from abc import ABC
from typing import Dict  # noqa
from typing import Optional  # noqa
from typing import TYPE_CHECKING
from typing import Union  # noqa

from slightly_better_types.function import Function

if TYPE_CHECKING:
    from typing import Generator


class AbstractBar(ABC):
    pass


class Bar(AbstractBar):
    pass


class Foo:
    def __init__(self):
        pass


class TestSlightlyBetterFunction(unittest.TestCase):
    integer: int = 1
    string: str = 'string'
    bar: Bar
    foo: Foo

    def true_values(self) -> "Generator":
        self.bar = Bar()

        # Arg test
        yield ['', [], {}]

        yield ['a: int', [self.integer], {}]
        yield ['a: str', [self.string], {}]
        yield ['*args: str', [self.string, self.string], {}]

        yield ['a: int, b: Union[int, str]', [self.integer, self.integer], {}]
        yield ['a: int, b: Union[int, str]', [self.integer, self.string], {}]

        yield ['a: Optional[int], b: Optional[int]', [None, None], {}]
        yield ['a: Optional[int], b: Optional[str]', [None, self.string], {}]

        yield ['a: dict', [{self.string: self.integer}], {}]

        # Kwarg Test
        yield [
            'a: int, b: Union[int, str], c: int = 2, *args, keyword_only: str, **kwargs',
            [],
            {'a': self.integer, 'b': self.string, 'keyword_only': 'word'}
        ]
        yield [
            'a: int, b: Union[int, str], c: int = 2, *args, keyword_only: str, **kwargs: str',
            [],
            {'a': self.integer, 'b': self.string, 'keyword_only': self.string, 'another_kwarg': self.string}
        ]

        # Arg and Kwarg Test
        yield [
            'a: int, b: Union[int, str], c: int = 2, *args, keyword_only: str, **kwargs',
            [self.integer],
            {'b': self.string, 'keyword_only': 'word'}
        ]

        yield [
            'a: int, b: Union[int, str], c: int = 2, *args, keyword_only: str, **kwargs: str',
            [self.integer, self.string],
            {'keyword_only': self.string, 'another_kwarg': self.string}
        ]

        # Inheritance
        yield ['bar: Bar, bar_2: AbstractBar', [self.bar, self.bar], {}]

    def false_values(self) -> "Generator":
        self.foo = Foo()

        yield ['a: str', [self.integer], {}]

        # Too many values
        yield ['', [self.string], {}]
        yield ['a: str', [self.string, self.string], {}]

        # Too few values
        yield ['a: str, b: str', [self.string], {}]

        # Wrong Args type
        yield ['*args: str', [self.integer, self.integer], {}]

        # Wrong kwargs type
        yield ['**args: str', [], {'keyword_arg': self.integer}]

        # Inheritance
        yield ['bar: Bar', [self.foo], {}]

    def test_true_values(self):
        for definition, args, kwargs in self.true_values():
            f = self._make_method(definition)
            assert f.accepts(*args, **kwargs)

    def test_false_values(self):
        for definition, args, kwargs in self.false_values():
            f = self._make_method(definition)
            assert not f.accepts(*args, *kwargs)

    def _make_method(self, definition: str) -> Function:
        exec(f"class TestClass:\n\tdef test(self,{definition}):\n\t\tpass")
        return Function(eval('TestClass.test'))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from inspect import Parameter as InspectParameter
from typing import Any

from izzet import izzet


class Parameter(InspectParameter):

    def __init__(self, type_hint, *args, **kwargs):
        super(Parameter, self).__init__(*args, **kwargs)
        self.accepted_type = type_hint

    @staticmethod
    def new(parameter: InspectParameter, type_hint: str) -> Parameter:
        return Parameter(
            name=parameter.name,
            kind=parameter.kind,
            default=parameter.default,
            annotation=parameter.annotation,
            type_hint=type_hint,
        )

    def get_name(self) -> str:
        return self.name

    def accepts(self, _input: Any) -> bool:
        return izzet(_input).a(self.accepted_type)



import os
import re
import sys

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 9)

if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write(f"""
==========================
Unsupported Python version
==========================
This version of Slightly Better Types requires Python {REQUIRED_PYTHON[0]}.{REQUIRED_PYTHON[1]}, but you're trying
to install it on Python {CURRENT_PYTHON[0]}.{CURRENT_PYTHON[1]}.
This may be because you are using a version of pip that doesn't
understand the python_requires classifier. Make sure you
have pip >= 9.0 and setuptools >= 24.2, then try again:
    $ python -m pip install --upgrade pip setuptools
    $ python -m pip install slightly-better-types
This will install the latest version of Slightly Better Types which works on
your version of Python. If you can't upgrade your pip (or Python), request
an older version of Slightly Better Types:
    $ python -m pip install "slightly-better-types<3.10"
""")
    sys.exit(1)


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version('slightly_better_types')

setuptools.setup(
    name="slightly-better-types",
    version=version,
    author="Peter Wensel",
    author_email="pwensel@gmail.com",
    url="https://github.com/slightly-better-python/slightly-better-types",
    description="Improved abstraction for dealing with union and named types.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],  # Information to filter the project on PyPi website
    python_requires='>=3.9',
    install_requires=['izzet>=1.0.2']
)
